from typing import Self

from anystore.util import join_uri
from pydantic import BaseModel, HttpUrl

from ftm_assets.settings import Settings
from ftm_assets.store import get_storage

settings = Settings()
storage = get_storage()

IMAGE_PREFIX = "img"


class Attribution(BaseModel):
    license: str
    license_url: HttpUrl
    author: str | None = None


class AltText(BaseModel):
    text: str
    language: str


class Image(BaseModel):
    id: str
    name: str
    url: HttpUrl
    alt: list[AltText] = []
    attribution: Attribution

    @property
    def key(self) -> str:
        return f"{IMAGE_PREFIX}/{self.id}/{self.name}"

    @property
    def thumbnail_key(self) -> str:
        return f"{IMAGE_PREFIX}/{self.id}/thumbs/{settings.thumbnail_size}.jpg"

    def get_public_url(self) -> HttpUrl:
        if settings.public_cdn_prefix is not None and storage.exists(self.key):
            return HttpUrl(join_uri(settings.public_cdn_prefix, self.key))
        return self.url

    def get_thumbnail_url(self) -> HttpUrl:
        if settings.public_cdn_prefix is not None and storage.exists(
            self.thumbnail_key
        ):
            return HttpUrl(join_uri(settings.public_cdn_prefix, self.thumbnail_key))
        return self.get_public_url()


class ApiImageResponse(Image):
    original_url: HttpUrl
    thumbnail_url: HttpUrl

    @classmethod
    def from_image(cls, image: Image) -> Self:
        return cls(
            id=image.id,
            name=image.name,
            alt=image.alt,
            attribution=image.attribution,
            url=image.get_public_url(),
            original_url=image.url,
            thumbnail_url=image.get_thumbnail_url(),
        )



from anystore.decorators import anycache, error_handler
from anystore.io import smart_open
from followthemoney.proxy import EntityProxy
from followthemoney.types import registry
from PIL import Image
from rigour.ids.wikidata import is_qid

from ftm_assets.logging import get_logger
from ftm_assets.model import Image as ImageModel
from ftm_assets.resolvers import wikidata
from ftm_assets.settings import Settings
from ftm_assets.store import get_cache, get_storage

settings = Settings()
log = get_logger(__name__)


@error_handler(logger=log)
@anycache(
    store=get_cache(),
    key_func=lambda i, s=None: f"thumbnails/{i.id}/{s or settings.thumbnail_size}",
)
def generate_thumbnail(img: "ImageModel", size: int | None = None) -> str:
    storage = get_storage()
    size = size or settings.thumbnail_size
    if not storage.exists(img.thumbnail_key):
        with smart_open(str(img.url)) as io:
            image = Image.open(io)
            rgb_img = image.convert("RGB")
            rgb_img.thumbnail((size, size))
            with storage.open(img.thumbnail_key, "wb") as out:
                rgb_img.save(out)
        log.info("Generated thumbnail.", image=img.id, size=size, uri=img.thumbnail_key)
    return img.thumbnail_key


@error_handler(logger=log)
@anycache(
    store=get_cache(),
    key_func=lambda i: f"mirrored/{i.id}",
)
def mirror(img: "ImageModel") -> str:
    storage = get_storage()
    if not storage.exists(img.key):
        with smart_open(str(img.url)) as io:
            with storage.open(img.key, "wb") as out:
                out.write(io.read())
        log.info("Stored image.", image=img.id, uri=img.key)
    return img.key


@error_handler(logger=log)
def lookup(
    id: str, store: bool | None = False, thumbnail: bool | None = False
) -> ImageModel | None:
    image = wikidata.resolve(id)
    if image is not None:
        if store or settings.mirror:
            mirror(image)
        if thumbnail or settings.thumbnails:
            generate_thumbnail(image)
        return image


@error_handler(logger=log)
def lookup_proxy(proxy: EntityProxy) -> ImageModel | None:
    id = str(proxy.id)
    if is_qid(id):
        return lookup(id)
    for id in proxy.get_type_values(registry.identifier):
        if is_qid(id):
            return lookup(id)



import secrets

from anystore.io import smart_read
from fastapi import Depends, FastAPI, Query
from fastapi.exceptions import HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import RedirectResponse

from ftm_assets import __version__
from ftm_assets.logic import lookup
from ftm_assets.model import ApiImageResponse
from ftm_assets.settings import Settings

settings = Settings()


def get_description() -> str:
    return smart_read(settings.api.description_uri)


AuthQuery = Query(
    None,
    description=(
        "Secret api key to allow assets mirroring and thumbnail generation "
        "(useful for e.g. static site builders)"
    ),
)


def get_authenticated(api_key: str = AuthQuery) -> bool:
    if not api_key:
        return False
    return secrets.compare_digest(api_key, settings.api.build_key)


DependsAuth = Depends(get_authenticated)


app = FastAPI(
    debug=settings.debug,
    title=settings.api.title,
    contact=settings.api.contact.model_dump(),
    description=get_description(),
    version=__version__,
    root_path=settings.api.path_prefix,
    redoc_url="/",
)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"] if settings.debug else settings.api.allowed_origins or ["*"],
    allow_methods=["OPTIONS", "GET"],
)

if settings.debug:
    from fastapi.staticfiles import StaticFiles

    app.mount("/static", StaticFiles(directory="static"), name="static")


@app.get("/img/{id}")
async def api_img_lookup(
    id: str, authenticated: bool = DependsAuth
) -> ApiImageResponse:
    """Get image metadata"""
    res = lookup(id, store=authenticated, thumbnail=authenticated)
    if res is None:
        raise HTTPException(status_code=404)
    return ApiImageResponse.from_image(res)


@app.get("/r/img/{id}")
async def api_img_redirect(id: str) -> RedirectResponse:
    """Get image redirect to public url"""
    res = lookup(id)
    if res is None:
        raise HTTPException(status_code=404)
    return RedirectResponse(url=res.get_public_url())


@app.get("/r/img/{id}/thumb")
async def api_img_thumbnail_redirect(id: str) -> RedirectResponse:
    """Get image redirect to public url"""
    res = lookup(id)
    if res is None:
        raise HTTPException(status_code=404)
    return RedirectResponse(url=res.get_thumbnail_url())



from anystore.model import StoreModel
from pydantic import BaseModel, Field  # , ImportString
from pydantic_settings import BaseSettings, SettingsConfigDict
from rich import print


class ApiContact(BaseModel):
    name: str = "Data and Research Center – DARC"
    url: str = "https://dataresearchcenter.org"
    email: str = "hi@dataresearchcenter.org"


class ApiSettings(BaseModel):
    path_prefix: str = "/api"
    """Deploy the fastapi under this prefix. This allows to have e.g. a domain
    "https://assets.example.org/api" for the app and
    "https://assets.example.org/*" deployed for static serving."""

    title: str = "FollowTheMoney Asset resolver"
    contact: ApiContact = ApiContact()
    description_uri: str = "README.md"
    allowed_origins: list[str] = []

    build_key: str = "secret-key-for-build"
    """Backend api key to use for build processes to allow mirror & thumbnail
    generation"""


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="ftm_assets_",
        env_nested_delimiter="_",
        nested_model_default_partial_update=True,
    )

    debug: bool = Field(default=False, alias="debug")
    """Debug mode"""

    cache: StoreModel = StoreModel(uri=".cache")
    """Lookup cache"""

    store: StoreModel = StoreModel(uri="static")
    """Storage for mirrored assets"""

    mirror: bool = False
    """Mirror assets from remotes on demand"""

    thumbnails: bool = False
    """Generate thumbnails on demand"""

    thumbnail_size: int = 600
    """Size (square) for generated thumbnails"""

    public_cdn_prefix: str | None = None
    """E.g. https://cdn.example.org"""

    # image_resolvers: list[ImportString] = ["ftm_assets.resolvers.wikidata:resolve"]
    # """Activated image resolvers"""

    api: ApiSettings = ApiSettings()


settings = Settings()

if settings.debug:
    print(settings)



"""
Wikidata resolver. Takes a wikidata id (QID) and finds the most recent image

https://www.wikidata.org/w/api.php?action=wbgetclaims&property=P18&entity=Q7747
"""

from datetime import datetime

import httpx
from anystore.decorators import anycache
from anystore.types import SDict
from banal import ensure_list
from pydantic import HttpUrl
from rigour.ids.wikidata import is_qid

from ftm_assets.logging import get_logger
from ftm_assets.model import Attribution, Image
from ftm_assets.settings import Settings
from ftm_assets.store import get_cache

log = get_logger(__name__)
settings = Settings()

BASE_URL = (
    "https://www.wikidata.org/w/api.php?action=wbgetclaims"
    "&property=P18&entity={qid}&format=json"
)

IMAGE_URL = (
    "https://commons.wikimedia.org/w/index.php?title=Special:Redirect/file/{name}"
)


def resolve_image_url(name: str) -> str | None:
    res = httpx.head(IMAGE_URL.format(name=name), follow_redirects=True)
    if res.status_code == 404:
        log.error("Image redirect not found", name=name)
        return
    res.raise_for_status()
    return str(res.url)


def extract_date(claim: SDict) -> str:
    for date in sorted(
        [
            p["datavalue"]["value"]["time"]
            for p in ensure_list(claim.get("qualifiers", {}).get("P585"))
        ],
        reverse=True,
    ):
        return date
    return datetime(1970, 1, 1).isoformat()


@anycache(
    store=get_cache(), key_func=lambda x: f"resolve/wikidata/{x}", ttl=3600, model=Image
)
def resolve(id: str) -> Image | None:
    # FIXME use `nomenklatura.wikidata` client?
    if is_qid(id):
        try:
            url = BASE_URL.format(qid=id)
            res = httpx.get(url)
            res.raise_for_status()
            data = res.json()
            candidates: list[SDict] = []
            for claim in ensure_list(data.get("claims", {}).get("P18")):
                candidates.append(
                    {
                        "name": claim["mainsnak"].get("datavalue", {}).get("value"),
                        "date": extract_date(claim),
                        "alt": [
                            p["datavalue"]["value"]
                            for p in ensure_list(
                                claim.get("qualifiers", {}).get("P2096")
                            )
                        ],
                    }
                )
            for candidate in sorted(candidates, key=lambda x: x["date"], reverse=True):
                url = resolve_image_url(candidate["name"])
                if url is not None:
                    return Image(
                        id=id,
                        name=candidate["name"],
                        url=HttpUrl(url),
                        alt=candidate["alt"],
                        attribution=Attribution(
                            license="CC BY 4.0",
                            license_url=HttpUrl(
                                "https://creativecommons.org/licenses/by/4.0/"
                            ),
                        ),
                    )
        except Exception as e:
            log.error(f"{e.__class__.__name__}: {e}")
            if settings.debug:
                raise e



"""
Task queue adapter for OpenAleph

For deferring jobs:
    queue: ftm-assets
    task: ftm_assets.tasks.resolve_image
"""

from openaleph_procrastinate.app import make_app
from openaleph_procrastinate.model import DatasetJob
from openaleph_procrastinate.tasks import task

from ftm_assets.logic import lookup_proxy

app = make_app(__loader__.name)


@task(app=app)
def resolve_image(job: DatasetJob) -> DatasetJob:
    lookup_proxy(job.entity)
    return job



from functools import cache

from anystore.store import BaseStore, get_store

from ftm_assets.settings import Settings

settings = Settings()


@cache
def get_storage() -> BaseStore:
    return get_store(**{**settings.store.model_dump(), "store_none_values": False})


@cache
def get_cache() -> BaseStore:
    return get_store(**{**settings.cache.model_dump(), "store_none_values": True})



from io import BytesIO
from typing import Annotated, Optional

import orjson
import typer
from anystore.cli import ErrorHandler
from anystore.io import smart_open, smart_stream, smart_stream_json_models
from ftmq.io import smart_stream_proxies
from rich import print

from ftm_assets import __version__, logic
from ftm_assets.logging import configure_logging
from ftm_assets.model import Image
from ftm_assets.settings import Settings

settings = Settings()

cli = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=settings.debug)


def write_obj(obj: Image, out: BytesIO) -> None:
    line = orjson.dumps(obj.model_dump(mode="json"), option=orjson.OPT_APPEND_NEWLINE)
    out.write(line)


@cli.callback(invoke_without_command=True)
def cli_ftm_assets(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False,
):
    if version:
        print(__version__)
        raise typer.Exit()
    configure_logging()


@cli.command()
def load_entities(
    input_uri: Annotated[
        str, typer.Option("-i", help="Input uri, default stdin")
    ] = "-",
    output_uri: Annotated[
        str, typer.Option("-o", help="Output uri, default stdout")
    ] = "-",
):
    with ErrorHandler():
        with smart_open(output_uri, "wb") as out:
            for proxy in smart_stream_proxies(input_uri):
                res = logic.lookup_proxy(proxy)
                if res is not None:
                    write_obj(res, out)


@cli.command()
def load_ids(
    input_uri: Annotated[
        str, typer.Option("-i", help="Input uri, default stdin")
    ] = "-",
    output_uri: Annotated[
        str, typer.Option("-o", help="Output uri, default stdout")
    ] = "-",
):
    with ErrorHandler():
        with smart_open(output_uri, "wb") as out:
            for id in smart_stream(input_uri, mode="r"):
                res = logic.lookup(id.strip())
                if res is not None:
                    write_obj(res, out)


@cli.command()
def mirror(
    input_uri: Annotated[
        str, typer.Option("-i", help="Input uri, default stdin")
    ] = "-",
    output_uri: Annotated[
        str, typer.Option("-o", help="Output uri, default stdout")
    ] = "-",
):
    """
    Mirror images to configured storage
    """
    with ErrorHandler():
        with smart_open(output_uri, "wb") as out:
            for image in smart_stream_json_models(input_uri, Image):
                logic.mirror(image)
                write_obj(image, out)


@cli.command()
def make_thumbnails(
    input_uri: Annotated[
        str, typer.Option("-i", help="Input uri, default stdin")
    ] = "-",
    output_uri: Annotated[
        str, typer.Option("-o", help="Output uri, default stdout")
    ] = "-",
    size: Annotated[int, typer.Option(help="Size in pixels")] = settings.thumbnail_size,
):
    """
    Generate thumbnails for given input results
    """
    with ErrorHandler():
        with smart_open(output_uri, "wb") as out:
            for image in smart_stream_json_models(input_uri, Image):
                logic.generate_thumbnail(image, size)
                write_obj(image, out)
